`timescale 1ns / 1ps
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Design Name: PPWM
// Module Name: ppwm
// Target Devices: UltraSCALE/UltraSCALE+
// Description: 
//   Distributed as an artifact of ‘A Power to Pulse Width Modulation Sensor for Remote Power Analysis Attacks’
//   Copyright (c) 2022 B. Udugama, D. Jayasinghe, H. Saadat, A. Ignjatovic and S. Parameswaran
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

module ppwm(
	input       resetn,
	input       clk,
	input       lbus_clk96,
	input       sampling_trig,
	input       uart_tx_Done,
	output reg  uart_tx_DV,
	output reg  uart_tx_Bit,
	output reg  busy
);

	reg		      ones_in_a_reading = 1'b1;
	reg		[7:0] readings_with_unmatching_ones;
	reg		[7:0] sampling_index;
	reg		[7:0] buffer_index;
	reg			  buffer_rden = 1'b0;
	reg			  buffer_wren = 1'b0;					  
	reg			  calib_sw = 1'b1;
	wire	      delayed_out;
	wire	      sampling_clk;
	wire	      ram_out;
	wire	      counter_out;

	parameter s_cap_IDLE        = 2'b00;
	parameter s_cap_FILLING     = 2'b01;
	parameter s_cap_FULL        = 2'b10;

	reg 	[1:0] r_FSM_cap = s_cap_IDLE; // FSM for handling data capturing

	parameter s_m_IDLE           = 3'b000;
	parameter s_m_SAMPLING       = 3'b001;
	parameter s_m_BUFFER_FULL    = 3'b010;
	parameter s_m_DELAY_CYCLE    = 3'b011;
	parameter s_m_TRANSFER       = 3'b100;
	parameter s_m_WAIT_FSM_TX    = 3'b101;
	parameter s_m_CLEANUP        = 3'b110;

	reg 	[2:0] r_FSM_main = s_m_IDLE; // The main FSM

	parameter s_tx_IDLE          = 2'b00;
	parameter s_tx_PREP          = 2'b01;
	parameter s_tx_WAITING       = 2'b10;
	parameter s_tx_DONE          = 2'b11;

	reg 	[1:0] r_FSM_tx = s_tx_IDLE; // FSM for transmitting the power trace
	
	// From idelay_fsm.v
	iodelay_fsm iodelay_fsm(
		.resetn(resetn),
		.clk(clk),
		.sampling_trig(sampling_trig),
		.CALIB_DIPSW(calib_sw),
		.lbus_clk96(lbus_clk96),
		.delayed_out(delayed_out),
		.undelayed_out(sampling_clk)
	);

	// Block RAM for storing the power trace
    xpm_memory_sdpram #(
	  .ADDR_WIDTH_A(8), 
	  .ADDR_WIDTH_B(8),
	  .CLOCKING_MODE("independent_clock"),
	  .MEMORY_OPTIMIZATION("false"),
	  .MEMORY_PRIMITIVE ("block"),
	  .MEMORY_SIZE(256),
	  .READ_DATA_WIDTH_B(1),
	  .READ_RESET_VALUE_B("0"),
      .READ_LATENCY_B(1),
	  .WRITE_DATA_WIDTH_A(1),
	  .BYTE_WRITE_WIDTH_A(1)
    ) buffer (
	  .rstb(~resetn), 
	  .dina(counter_out),
	  .addra(sampling_index),
	  .clka(sampling_clk),
	  .ena(buffer_wren),
	  .wea(1'b1),
	  .doutb(ram_out),
	  .addrb(buffer_index),
	  .clkb(clk),
	  .enb(buffer_rden)
    );


	// From gate_net.v
	gate_net gate_net (
		.clk(sampling_clk),
		.delayed_in(delayed_out),
		.undelayed_in(sampling_clk),
		.counter_value(counter_out)
	); 
 
	always @(posedge sampling_clk)
	begin
		case (r_FSM_cap)
			s_cap_IDLE :
			begin
				sampling_index <= 9'b0;
				if ((r_FSM_main == s_m_SAMPLING) & (resetn == 1'b1))
				begin
					buffer_wren <= 1'b1;
					r_FSM_cap <= s_cap_FILLING;
				end
			end

			s_cap_FILLING :
			begin
				sampling_index <= sampling_index + 1'b1;
				if (sampling_index == 255)
					r_FSM_cap <= s_cap_FULL;
			end

			s_cap_FULL :
			begin
				buffer_wren <= 1'b0;
				if (r_FSM_main == s_m_BUFFER_FULL)
					r_FSM_cap <= s_cap_IDLE;
			end

		default :
			r_FSM_cap <= s_cap_IDLE;
		endcase
	end

	always @(posedge clk)
	begin  
		case (r_FSM_main)
			s_m_IDLE :
			begin
				if ((sampling_trig == 1'b1) & (resetn == 1'b1))
				begin
					busy <= 1'b1;
					r_FSM_main <= s_m_SAMPLING;
				end
				else
				begin
					busy <= 1'b0;
					buffer_rden <= 0;
				end
			end

			s_m_SAMPLING :
			begin
				if (r_FSM_cap == s_cap_FULL)
				begin
					buffer_rden <= 1; 
					buffer_index <= 9'b0;
					r_FSM_main <= s_m_BUFFER_FULL;
				end
			end

			s_m_BUFFER_FULL :
			begin
				readings_with_unmatching_ones <= 0;
				r_FSM_main <= s_m_DELAY_CYCLE;
			end
			 
			s_m_DELAY_CYCLE :
			begin
				uart_tx_Bit <= ram_out;
				if ((buffer_index > 144) && (buffer_index < 177)) // Bounds of the samples of interest
				begin
					if(ram_out != ones_in_a_reading) // A mismatch in the number of set bits in the sample and the number of set bits stored in ones_in_a_reading
						readings_with_unmatching_ones <= readings_with_unmatching_ones + 1; // Increment the readings_with_unmatching_ones ~ decrementing 'odds' variable
				end
				r_FSM_main <= s_m_TRANSFER;
			end

			s_m_TRANSFER :
			begin
				if (r_FSM_tx == s_tx_IDLE)
					r_FSM_main <= s_m_WAIT_FSM_TX;
			end

			s_m_WAIT_FSM_TX :
			begin
				if (r_FSM_tx == s_tx_DONE)
				begin 
					if (buffer_index == 255)
					begin
						if (readings_with_unmatching_ones == 32) // None of the readings were matched -> ones_in_a_reading was wrong. Execute again with a new ones_in_a_reading value
							ones_in_a_reading <= ram_out;
						else if ((readings_with_unmatching_ones < 32) && (readings_with_unmatching_ones > 0))
							calib_sw <= 1'b0; // Turn off calibration.
						r_FSM_main <= s_m_CLEANUP;
					end
					else
					begin
						buffer_index <= buffer_index + 1'b1;
						r_FSM_main <= s_m_DELAY_CYCLE;
					end
				end
			end

			s_m_CLEANUP :
			begin
				busy <= 1'b0;
				buffer_rden <= 0;
				r_FSM_main <= s_m_IDLE;
			end

		default :
			r_FSM_main <= s_m_IDLE;

		endcase
	end

	always @(posedge clk)
	begin  
		case (r_FSM_tx)
			s_tx_IDLE :
			begin
				uart_tx_DV <= 1'b0;
				if (r_FSM_main == s_m_WAIT_FSM_TX)
					r_FSM_tx <= s_tx_PREP;
			end

			s_tx_PREP :
			begin
				uart_tx_DV <= 1'b1;
				r_FSM_tx <= s_tx_WAITING;
			end

			s_tx_WAITING :
			begin
				uart_tx_DV <= 1'b0;
				if (uart_tx_Done == 1'b1)
					r_FSM_tx <= s_tx_DONE;
			end

			s_tx_DONE :
			begin
				if ((r_FSM_main == s_m_CLEANUP) | (r_FSM_main == s_m_TRANSFER))
					r_FSM_tx <= s_tx_IDLE;
			end

		default :
			r_FSM_tx <= s_tx_IDLE;

		endcase
	end
	 
endmodule